use api::game_actions::SpawnToken;
use game::{chat_message::ChatMessage, entry::{ActionDefinition, DiceRoll}, Game, GameImpl};
use tokio::sync::{broadcast, mpsc};

pub mod api;
pub mod game;
pub mod pathfinder2r_impl;
pub mod table;
pub mod user;

pub struct GameServer {
    _game: Game<pathfinder2r_impl::Pathfinder2rCharacterSheet, pathfinder2r_impl::entry::Entry>,
    tokens: Vec<(String, i32, i32)>,
    chat: Vec<(String, ChatMessage)>,
}
impl GameServer {
    pub fn new() -> Self {
        Self {
            _game: Game::new(),
            tokens: vec![("assets/pf2r/tokens/louise.jpg".to_string(), 2, 2)],
            chat: vec![
                (
                    "Server".to_string(), 
                    ChatMessage::new("a weapon description".to_string())
                        .id(1)
                        .character(Some("Sword or something".to_string()))
                        .with_action(
                            ActionDefinition::new("weapon/attack".to_string())
                            .display_name(Some("Attack +7".to_string()))
                            .with_roll(DiceRoll::new("Pierce".to_string(), 12, 1).constant(1))
                            .with_roll(DiceRoll::new("Fire".to_string(), 4, 2))
                        )
                        .with_action(ActionDefinition::new("Attack +3".to_string()))
                        .with_action(ActionDefinition::new("Attack -1".to_string()))
                )
            ],
        }
    }

    pub async fn server_loop(
        mut self,
        mut msgs: mpsc::Receiver<(String, api::Request)>,
        broadcast: broadcast::Sender<(Option<String>, api::Response)>,
    ) {
        while let Some(req) = msgs.recv().await {
            // TODO: do stuff yo!
            let (id, req) = req;
            println!("Got message from {}: {:?}", &id, &req);

            match req {
                api::Request::Error => {}
                api::Request::Login(_) => {}
                api::Request::Message(mut msg) => {
                    if msg.id == 0 || msg.id > self.chat.len() {
                        msg.id = self.chat.len() + 1; // set the message id, 0 is invalid
                    }
                    // TODO: check if the editor is an admin as well
                    else if id == self.chat[msg.id - 1].0 {
                        self.chat[msg.id - 1] = (id.clone(), msg.clone());
                    } else {
                        // if its an edit message and editor is not the owner, skip
                        continue;
                    }
                    // Force the sender id to be the new id of the message, even if an id was provided
                    msg.source = id.clone();

                    self.chat.push((id.clone(), msg.clone()));
                    if msg.whisper.is_some() {
                        _ = broadcast.send((Some(id.clone()), api::Response::Message(msg.clone())));
                    }
                    _ = broadcast.send((msg.whisper.clone(), api::Response::Message(msg)));
                }
                api::Request::GetChatHistory {
                    mut amount,
                    from: last_msg,
                } => {
                    if amount == 0 {
                        amount = self.chat.len();
                    }
                    let history: Vec<ChatMessage> = self
                        .chat
                        .iter()
                        .skip(last_msg)
                        .take(amount)
                        .map(|m| m.1.clone())
                        .collect();
                    _ = broadcast.send((Some(id), api::Response::GetChatHistory(history)));
                }
                api::Request::GetLastMessages { mut amount } => {
                    if amount == 0 {
                        amount = self.chat.len();
                    }
                    let start = if amount >= self.chat.len() {
                        0
                    } else {
                        self.chat.len() - amount
                    };
                    let history: Vec<ChatMessage> =
                        self.chat.iter().skip(start).map(|m| m.1.clone()).collect();
                    _ = broadcast.send((Some(id), api::Response::GetChatHistory(history)));
                },
                api::Request::GetTokens => {
                    for (i, (path, x, y)) in self.tokens.iter().enumerate() {
                        let bits = std::fs::read(path).expect("FAILED READING TOKEN IMAGE");
                        let img = base64::Engine::encode(&base64::prelude::BASE64_STANDARD, &bits);
                        _ = broadcast.send((Some(id.clone()), api::Response::SpawnToken(SpawnToken { token_id: i, x: *x, y: *y, img })));
                    }
                },
                api::Request::SpawnToken { x, y, img_path } => {
                    let token_id = self.tokens.len();
                    self.tokens.push((img_path.clone(), x, y));
                    let bits = std::fs::read(img_path).expect("FAILED READING TOKEN IMAGE");
                    let img = base64::Engine::encode(&base64::prelude::BASE64_STANDARD, &bits);
                        _ = broadcast.send((Some(id.clone()), api::Response::SpawnToken(SpawnToken { token_id, x, y, img })));
                },
                api::Request::MoveToken { token_id, x, y } => {
                    // TODO: add check to make sure the actor is authorized to move the token
                    if token_id < self.tokens.len() {
                        self.tokens[token_id].1 = x;
                        self.tokens[token_id].2 = y;
                        _ = broadcast.send((None, api::Response::MoveToken { token_id, x, y }));
                    }
                },
                api::Request::ActionResult(result) => {
                    let msg = ChatMessage::new(
                        result.results.iter()
                            // .map(|d| &d.result_text)
                            .fold(String::new(), |a,b| a + &format!("{}: {} = {}\n", &b.name, &b.result_text, b.result))
                        )
                        .character(Some(result.name))
                        .source(id.clone())
                        .targets(Some(result.targets))
                        .id(self.chat.len() + 1);
                    self.chat.push((id, msg.clone()));
                    _ = broadcast.send((None, api::Response::Message(msg)));
                        
                }
                api::Request::Quit => _ = broadcast.send((None, api::Response::Quit { id })),
                api::Request::Kick(id) => _ = broadcast.send((Some(id), api::Response::Shutdown)),
                api::Request::Shutdown => break,
            }
        }
        _ = broadcast.send((None, api::Response::Shutdown));
    }
}
